import timeit

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from newton_type import newton_type


# for people who cannot see the plots
if mpl.get_backend() == 'agg':
    mpl.use('WebAgg')
print(f"backend: {mpl.get_backend()}")


## plot the objective function

# define the function as anonymous function handle
# you can then evaluate it as f(x,y) (elementwise)
f = lambda x,y: .5*((x-1)**2 + 100*(y-x**2)**2 + y**2)

# Create circle grid of x y values
dR = .002
N = int(np.sqrt(2) / dR)
R = np.linspace(0, np.sqrt(2), N)
TH = 2*np.pi*np.linspace(0, 1, 500) 
X = np.outer(R.T, np.cos(TH)) 
Y = np.outer(R.T, np.sin(TH))
# calculate objective function values
# add 1 and take log for better visualization
Z = np.log(1 + f(X,Y))

fig = plt.figure(1)


# subplot 1 with surface plot
ax = fig.add_subplot(1,2,1, projection='3d')
ax.view_init(100, 50)
ax.grid(True)
# ax.axis('square')  # NOTE: cannot scale Axes3D
ax.plot_surface(X,Y,Z, linewidth=0, cmap='coolwarm')
ax.set_xlabel(r'$x$')
ax.set_ylabel(r'$y$')
ax.set_zlabel(r'$\log(1+f(x,y))$')
# read current axis limits and fix them at these values
ax_lim = ax.axis()
ax.axis(ax_lim)

# subplot 2 with contour plot
ax = fig.add_subplot(1,2,2, projection='3d')
ax.view_init(100, 50)
ax.grid(True)
# ax.axis('square')  # NOTE: cannot scale Axes3D
ax.contour(X,Y,Z, cmap='gray')
ax.set_xlabel(r'$x$')
ax.set_ylabel(r'$y$')
ax.set_zlabel(r'$\log(1+f(x,y))$')
# fix the axis limits at the same values
ax.axis(ax_lim)

fig.suptitle(r'Surface and contour plot of the nonlinear function')


## grad and hessians

# gradient
fgrad = lambda x,y: [
    x - 1 - 200*x*y + 200*x**3,
    101*y - 100*x**2
]

# true hessian
fhess = lambda x,y: [
    [1 - 200*y + 600*x**2, -200*x],
    [               -200*x,  101  ]
]

# Gauss newton hessian
B_GN = lambda x,y: [
    [1 + 400*x**2, -200*x],
    [       -200*x,  101  ]
]
            
z0 = [-1, 1]                            # initial guesss
tol = 1e-3                             # tolerance for stopping criterion
maxit = 5000
## newton Type algorithms
# introduce z = [xy]
fgradz = lambda z: fgrad(z[0], z[1])
# exact hessian
t_ref = timeit.default_timer()
Z_ex, grad_ex = newton_type(fgradz, lambda z: fhess(z[0], z[1]), z0, tol, maxit)
t_ex = timeit.default_timer() - t_ref
# Gauss Newton
t_ref = timeit.default_timer()
Z_GN, grad_GN = newton_type(fgradz, lambda z: B_GN(z[0], z[1]), z0, tol, maxit)
t_GN = timeit.default_timer() - t_ref
# three versions of gradient descent
t_ref = timeit.default_timer()
Z_gd1, grad_gd1 = newton_type(fgradz, lambda z: 100*np.eye(2), z0, tol, maxit)
t_gd1 = timeit.default_timer() - t_ref
t_ref = timeit.default_timer()
Z_gd2, grad_gd2 = newton_type(fgradz, lambda z: 200*np.eye(2), z0, tol, maxit)
t_gd2 = timeit.default_timer() - t_ref
t_ref = timeit.default_timer()
Z_gd5, grad_gd5 = newton_type(fgradz, lambda z: 500*np.eye(2), z0, tol, maxit)
t_gd5 = timeit.default_timer() - t_ref

## print infos
print(
    f"\n"
    f"------------------------------------\n"
    f"Method\t\t  #iter\t  #t in s\n"
    f"------------------------------------\n"
    f"exact hessian     #{len(Z_ex)-1:4d}   #{t_ex:.3e}\n"
    f"Gauss-Newton      #{len(Z_GN)-1:4d}   #{t_GN:.3e}\n"
    f"grad desc, a=100  #{len(Z_gd1)-1:4d}   #{t_gd1:.3e}\n"
    f"grad desc, a=200  #{len(Z_gd2)-1:4d}   #{t_gd2:.3e}\n"
    f"grad desc, a=500  #{len(Z_gd5)-1:4d}   #{t_gd5:.3e}\n"
)

## add iterates to plot
fig.add_subplot(ax)
x = np.linspace(ax_lim[0], ax_lim[1], 100)
ax.plot(x, x**2,'m--', label=r'$x=y^2$')
ax.plot(Z_ex[0], Z_ex[1], 'b-x', label=r'exact Hessian')
ax.plot(Z_GN[0], Z_GN[1], 'r-x', label=r'GN')
ax.plot(Z_gd5[0], Z_gd5[1], 'g-x', label=r'GD, $\alpha=500$')
ax.plot(Z_gd2[0], Z_gd2[1], 'k-x', label=r'GD, $\alpha=200$')
ax.plot(Z_gd1[0], Z_gd1[1], 'c-x', label=r'GD, $\alpha=100$')
leg = ax.legend(loc='lower right')


## show plots
plt.show()
